import argparse
import json
import logging
import os
import shutil
import glob

import numpy as np
import torch
from torch.utils.data import DataLoader, RandomSampler, SequentialSampler
from fastprogress.fastprogress import master_bar, progress_bar
from attrdict import AttrDict

from transformers import (
    AutoConfig,
    AutoModelForSequenceClassification,
    AutoTokenizer,
    AdamW,
    get_linear_schedule_with_warmup,
)

from src import init_logger, set_seed, compute_metrics
from processor import seq_cls_load_and_cache_examples as load_and_cache_examples
from processor import seq_cls_processors as processors
from processor import seq_cls_output_modes as output_modes

logger = logging.getLogger(__name__)


def train(
    args, model, train_dataset, dev_dataset=None, test_dataset=None, tokenizer=None
):
    train_sampler = RandomSampler(train_dataset)
    train_dataloader = DataLoader(
        train_dataset, sampler=train_sampler, batch_size=args.train_batch_size
    )
    if args.max_steps > 0:
        t_total = args.max_steps
        args.num_train_epochs = (
            args.max_steps
            // (len(train_dataloader) // args.gradient_accumulation_steps)
            + 1
        )
    else:
        t_total = (
            len(train_dataloader)
            // args.gradient_accumulation_steps
            * args.num_train_epochs
        )

    # Prepare optimizer and schedule (linear warmup and decay)
    no_decay = ["bias", "LayerNorm.weight"]
    optimizer_grouped_parameters = [
        {
            "params": [
                p
                for n, p in model.named_parameters()
                if not any(nd in n for nd in no_decay)
            ],
            "weight_decay": args.weight_decay,
        },
        {
            "params": [
                p
                for n, p in model.named_parameters()
                if any(nd in n for nd in no_decay)
            ],
            "weight_decay": 0.0,
        },
    ]
    optimizer = AdamW(
        optimizer_grouped_parameters, lr=args.learning_rate, eps=args.adam_epsilon
    )
    scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=int(t_total * args.warmup_proportion),
        num_training_steps=t_total,
    )

    if os.path.isfile(os.path.join(args.model_dir, "optimizer.pt")) and os.path.isfile(
        os.path.join(args.model_dir, "scheduler.pt")
    ):
        # Load optimizer and scheduler states
        optimizer.load_state_dict(
            torch.load(os.path.join(args.model_dir, "optimizer.pt"))
        )
        scheduler.load_state_dict(
            torch.load(os.path.join(args.model_dir, "scheduler.pt"))
        )

    # Train!
    logger.info("***** Running training *****")
    logger.info("  Num examples = %d", len(train_dataset))
    logger.info("  Num Epochs = %d", args.num_train_epochs)
    logger.info("  Total train batch size = %d", args.train_batch_size)
    logger.info("  Gradient Accumulation steps = %d", args.gradient_accumulation_steps)
    logger.info("  Total optimization steps = %d", t_total)
    logger.info("  Logging steps = %d", args.logging_steps)
    logger.info("  Save steps = %d", args.save_steps)

    global_step = 0
    tr_loss = 0.0

    model.zero_grad()
    mb = master_bar(range(int(args.num_train_epochs)))
    for epoch in mb:
        epoch_iterator = progress_bar(train_dataloader, parent=mb)
        for step, batch in enumerate(epoch_iterator):
            model.train()
            batch = tuple(t.to(args.device) for t in batch)
            inputs = {
                "input_ids": batch[0],
                "attention_mask": batch[1],
                "labels": batch[3],
            }
            if model.config.model_type not in ["roberta"]:
                inputs["token_type_ids"] = batch[2]
            outputs = model(**inputs)

            loss = outputs[0]

            if args.gradient_accumulation_steps > 1:
                loss = loss / args.gradient_accumulation_steps

            loss.backward()
            tr_loss += loss.item()
            if (step + 1) % args.gradient_accumulation_steps == 0 or (
                len(train_dataloader) <= args.gradient_accumulation_steps
                and (step + 1) == len(train_dataloader)
            ):
                torch.nn.utils.clip_grad_norm_(model.parameters(), args.max_grad_norm)

                optimizer.step()
                scheduler.step()
                model.zero_grad()
                global_step += 1

                if args.logging_steps > 0 and global_step % args.logging_steps == 0:
                    if args.evaluate_test_during_training:
                        evaluate(args, model, test_dataset, "test", global_step)
                    else:
                        evaluate(args, model, dev_dataset, "dev", global_step)

                if args.save_steps > 0 and global_step % args.save_steps == 0:
                    # Save model checkpoint
                    output_dir = os.path.join(
                        args.output_dir, "checkpoint-{}".format(global_step)
                    )
                    if not os.path.exists(output_dir):
                        os.makedirs(output_dir)
                    model_to_save = model.module if hasattr(model, "module") else model
                    model_to_save.save_pretrained(output_dir)
                    if tokenizer:
                        tokenizer.save_pretrained(output_dir)
                        logger.info("Saving tokenzer to {}".format(output_dir))

                    torch.save(args, os.path.join(output_dir, "training_args.bin"))
                    logger.info("Saving model checkpoint to {}".format(output_dir))

                    if args.save_optimizer:
                        torch.save(
                            optimizer.state_dict(),
                            os.path.join(output_dir, "optimizer.pt"),
                        )
                        torch.save(
                            scheduler.state_dict(),
                            os.path.join(output_dir, "scheduler.pt"),
                        )
                        logger.info(
                            "Saving optimizer and scheduler states to {}".format(
                                output_dir
                            )
                        )

            if args.max_steps > 0 and global_step > args.max_steps:
                break

        mb.write("Epoch {} done".format(epoch + 1))

        if args.max_steps > 0 and global_step > args.max_steps:
            break

    return global_step, tr_loss / global_step


def evaluate(args, model, eval_dataset, mode, global_step=None):
    results = {}
    eval_sampler = SequentialSampler(eval_dataset)
    eval_dataloader = DataLoader(
        eval_dataset, sampler=eval_sampler, batch_size=args.eval_batch_size
    )

    # Eval!
    if global_step is not None:
        logger.info(
            "***** Running evaluation on {} dataset ({} step) *****".format(
                mode, global_step
            )
        )
    else:
        logger.info("***** Running evaluation on {} dataset *****".format(mode))
    logger.info("  Num examples = {}".format(len(eval_dataset)))
    logger.info("  Eval Batch size = {}".format(args.eval_batch_size))
    eval_loss = 0.0
    nb_eval_steps = 0
    preds = None
    out_label_ids = None

    for batch in progress_bar(eval_dataloader):
        model.eval()
        batch = tuple(t.to(args.device) for t in batch)

        with torch.no_grad():
            inputs = {
                "input_ids": batch[0],
                "attention_mask": batch[1],
                "labels": batch[3],
            }
            if model.config.model_type not in ["roberta"]:
                inputs["token_type_ids"] = batch[2]
            outputs = model(**inputs)
            tmp_eval_loss, logits = outputs[:2]

            eval_loss += tmp_eval_loss.mean().item()
        nb_eval_steps += 1
        if preds is None:
            preds = logits.detach().cpu().numpy()
            out_label_ids = inputs["labels"].detach().cpu().numpy()
        else:
            preds = np.append(preds, logits.detach().cpu().numpy(), axis=0)
            out_label_ids = np.append(
                out_label_ids, inputs["labels"].detach().cpu().numpy(), axis=0
            )

    eval_loss = eval_loss / nb_eval_steps
    if output_modes[cli_args.data] == "classification":
        preds = np.argmax(preds, axis=1)
    elif output_modes[cli_args.data] == "regression":
        preds = np.squeeze(preds)
    result = compute_metrics(cli_args.task, out_label_ids, preds)
    model_prediction = [out_label_ids, preds]
    results.update(result)

    if args.do_train:
        output_dir = os.path.join(args.output_dir, mode)
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        output_eval_file = os.path.join(
            output_dir,
            "{}-{}.txt".format(mode, global_step)
            if global_step
            else "{}.txt".format(mode),
        )
        with open(output_eval_file, "w") as f_w:
            for key in sorted(results.keys()):
                f_w.write("  {} = {}\n".format(key, str(results[key])))

    logger.info("***** Eval results on {} dataset *****".format(mode))
    for key in sorted(results.keys()):
        logger.info("  {} = {}".format(key, str(results[key])))

    return results, model_prediction


def main(cli_args):
    # Read from config file and make args
    with open(os.path.join(cli_args.config_dir, cli_args.data, f"{cli_args.model_size}.json")) as f:
        args = AttrDict(json.load(f))
    logger.info("Training/evaluation parameters {}".format(args))

    args.model_dir = cli_args.model_dir
    args.data = cli_args.data
    args.task = cli_args.task

    if cli_args.learning_rate is not None:
        args.learning_rate = cli_args.learning_rate

    if cli_args.version is not None:
        args.output_dir = os.path.join(args.ckpt_dir, f"{args.task}-{cli_args.version}")
    else:
        args.output_dir = os.path.join(args.ckpt_dir, args.task)

    init_logger()
    set_seed(args)

    processor = processors[cli_args.data](args)
    labels = processor.get_labels()
    if output_modes[cli_args.data] == "regression":
        config = AutoConfig.from_pretrained(
            args.model_dir,
            num_labels=processor.get_num_labels(),
        )
    else:
        config = AutoConfig.from_pretrained(
            args.model_dir,
            num_labels=processor.get_num_labels(),
            id2label={str(i): label for i, label in enumerate(labels)},
            label2id={label: i for i, label in enumerate(labels)},
        )
    tokenizer = AutoTokenizer.from_pretrained(
        args.model_dir,
        do_lower_case=args.do_lower_case,
        use_fast=False,
    )
    model = AutoModelForSequenceClassification.from_pretrained(
        args.model_dir, config=config
    )

    # GPU or CPU
    args.device = "cuda" if torch.cuda.is_available() and not args.no_cuda else "cpu"
    model.to(args.device)

    # Load dataset
    if args.do_train:
        train_dataset = (
            load_and_cache_examples(args, tokenizer, mode="train")
            if args.train_file
            else None
        )
        dev_dataset = (
            load_and_cache_examples(args, tokenizer, mode="dev")
            if args.dev_file
            else None
        )
        if dev_dataset is None:
            args.evaluate_test_during_training = (
                True  # If there is no dev dataset, only use testset
            )

    test_dataset = (
        load_and_cache_examples(args, tokenizer, mode="test")
        if args.test_file
        else None
    )

    if args.do_train:
        global_step, tr_loss = train(
            args, model, train_dataset, dev_dataset, test_dataset, tokenizer
        )
        logger.info(" global_step = {}, average loss = {}".format(global_step, tr_loss))

    results = {}
    id2label = {str(i): label for i, label in enumerate(labels)}

    if args.do_eval:
        best_result, best_step = 0, 0
        best_model_prediction = list()
        checkpoints = list(
            os.path.dirname(c)
            for c in sorted(
                glob.glob(
                    args.output_dir + "/**/" + "pytorch_model.bin", recursive=True
                )
            )
        )
        if not args.eval_all_checkpoints:
            checkpoints = checkpoints[-1:]
        else:
            logging.getLogger("transformers.configuration_utils").setLevel(
                logging.WARN
            )  # Reduce logging
            logging.getLogger("transformers.modeling_utils").setLevel(
                logging.WARN
            )  # Reduce logging
        logger.info("Evaluate the following checkpoints: %s", checkpoints)

        for checkpoint in checkpoints:
            global_step = checkpoint.split("-")[-1]
            model = AutoModelForSequenceClassification.from_pretrained(checkpoint)
            model.to(args.device)
            result, model_prediction = evaluate(
                args, model, test_dataset, mode="test", global_step=global_step
            )
            for k, v in result.items():
                if k == args.metric:
                    if best_result < v:
                        best_result_dict = result
                        best_result = v
                        best_step = global_step
                        best_model_prediction = model_prediction
            result = dict((k + "_{}".format(global_step), v) for k, v in result.items())
            results.update(result)

        # Write Best Result
        best_result_file = os.path.join(args.output_dir, "best_result.txt")
        with open(best_result_file, "w") as f_w:
            f_w.write(f"Step {best_step}, ")
            for key in sorted(best_result_dict.keys()):
                f_w.write("{} = {:.4f}\t".format(key, best_result_dict[key]))

        output_eval_file = os.path.join(args.output_dir, "eval_results.txt")
        with open(output_eval_file, "w") as f_w:
            for key in sorted(results.keys()):
                f_w.write("{} = {}\n".format(key, str(results[key])))

        # [boychaboy] write best model prediction
        model_prediction_file = os.path.join(args.output_dir, "test_prediction.tsv")
        with open(model_prediction_file, "w", encoding="utf-8") as f_w:
            test_examples = processor.get_examples("test")
            texts = [example.text_a for example in test_examples]
            titles = [example.text_b for example in test_examples]
            f_w.write("text\tgold\tpred\n")
            for text, title, gold, pred in zip(
                texts, titles, best_model_prediction[0], best_model_prediction[1]
            ):
                f_w.write(f"{text}\t{id2label[str(gold)]}\t{id2label[str(pred)]}\n")

    if args.do_predict:
        if not os.path.exists(cli_args.predict_dir):
            os.makedirs(cli_args.predict_dir)
        logger.info("Predicting with the following model: %s", args.model_dir)
        result, model_prediction = evaluate(args, model, test_dataset, mode="test")
        result_file = os.path.join(cli_args.predict_dir, "result.txt")
        with open(result_file, "w") as f_w:
            f_w.write("Step {best_step}, ")
            for key in sorted(result.keys()):
                f_w.write("{} = {:.4f}\t".format(key, result[key]))
        logger.info("Test result written in : %s", result_file)
        model_prediction_file = os.path.join(
            cli_args.predict_dir, "test_prediction.tsv"
        )
        logger.info("Model prediction written in : %s", model_prediction_file)

        with open(model_prediction_file, "w", encoding="utf-8") as f_w:
            test_examples = processor.get_examples("test")
            texts = [example.text_a for example in test_examples]
            titles = [example.text_b for example in test_examples]
            f_w.write("text\tgold\tpred\n")
            for text, title, gold, pred in zip(
                texts, titles, model_prediction[0], model_prediction[1]
            ):
                f_w.write(f"{text}\t{id2label[str(gold)]}\t{id2label[str(pred)]}\n")

    # Delete Checkpoints
    directories = glob.glob("{}/checkpoint-*/".format(args.output_dir))
    for directory in directories:
        directory_step = directory[:-1].split("-")[-1]
        if int(best_step) != int(directory_step):
            shutil.rmtree(directory)


if __name__ == "__main__":
    cli_parser = argparse.ArgumentParser()

    cli_parser.add_argument("--task", type=str, required=True)
    cli_parser.add_argument("--data", type=str, required=True)
    cli_parser.add_argument("--config_dir", type=str, default="config")
    cli_parser.add_argument(
        "--model_size", type=str, required=True, help="base, small etc"
    )
    cli_parser.add_argument("--model_dir", type=str, required=True)
    cli_parser.add_argument("--learning_rate", type=float)
    cli_parser.add_argument("--version", type=str)
    cli_parser.add_argument(
        "--predict_dir", type=str
    )

    cli_args = cli_parser.parse_args()

    main(cli_args)
